/*
 * Copyright (c) 1993-2010 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/**
 ** gyroServerLib.c
 **
 ** Functions to open, read and decode gyro data on a serial line :
 **
 **      int gyroOpen(char *device)
 **      int gyroRead(int fd, char *buf, u_long *date)
 **      int gyroDecode(GyroData *gyro)
 **
 ** Author: Anthony Mallet (original gyro.c) and Sara Fleury
 ** Date: 
 **
 **/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef HAS_POSIX_CLOCK
#include <time.h>
#else
#include <sys/time.h>
#endif

#include <portLib.h>
#include <h2mathLib.h>
#include "serial.h"

#include "gyroKVH1000HardLib.h"

static u_long getTime (void);

/*
 * gyroOpen  -------------------------------------------------------------
 *
 * Open connection to serial port
 *
 * Returns : fd
 */

int gyroOpen(char *device)
{
  int fd;

   fd = openSerial(device, GYRO_SERIAL_SPEED);
   if (fd < 0) {
     printf ("gyroOpen: openSerial (%s) failed errno = %d\n", device, errno);
     return -1;
   }
   return fd;
}


/*
 * gyroRead -------------------------------------------------------------
 *
 * Read a message
 */

int gyroRead(int fd, char *buf, u_long *msec)
{
   int i;

   /* read 8 bytes */
   for(i=0; i<8; i++) {
	   if (read(fd, &buf[i], 1) <= 0) {
		   fprintf(stderr, "gyroRead: error read %s\n",
		       strerror(errno));
		   return -1;
	   }
   }

  sync:
   /* continue read until a message header is found (8th bit of first byte
    * must be set) */
   while(!(buf[0] & 0x80)) {
      printf("gyroRead: skipping one byte\n");

      memcpy(buf, buf+1, 7);
      if (read(fd, buf+7, 1) <= 0) return -1;
   }

   /* check message is complete (8th bit of the 7 last bytes must be
    * unset) */
   for(i=1; i<8; i++) if (buf[i] & 0x80) {
      printf("gyroRead: incomplete message\n");

      memcpy(buf, buf+1, 7);
      goto sync;
   }

   *msec = getTime();

   return 0;
}


/*
 * gyroDecode -----------------------------------------------------------
 *
 * Decode a raw message
 */

int gyroDecode(GyroData *gyro)
{
   int d;
   char *buf = gyro->buf;
   unsigned char c;
   static int totMeasuresOffset=0;

   /* check checksum */
   c  = (buf[0]<<1) + ((buf[1]&0x7F)>>6);
   c += (buf[1]<<2) + ((buf[2]&0x7F)>>5);
   c += (buf[2]<<3) + ((buf[3]&0x7F)>>4);
   c += (buf[3]<<4) + ((buf[4]&0x7F)>>3);
   c += (buf[4]<<5) + ((buf[5]&0x7F)>>2);
   c += (buf[5]<<6) + ((buf[6]&0x7F)>>1);
   c += (buf[6]<<7) + (buf[7]&0x7F);
   if (c) {
      fputs("gyro: bad checksum\n", stderr);
      return -1;
   }

   /* check "built-in-test" bit */
   if (!(buf[3] & 0x20)) {
      fputs("gyro: built-in-test failed\n", stderr);
      return -1;
   }

   /* serial number */
   if (!gyro->snO) {
      if (buf[3] & 0x40) {
	 if (!gyro->snI) {
	    /* first bit */
	    gyro->snBit = 1;
	    gyro->snI = 1;
	    gyro->sn = 0;
	 } else {
	    /* last bit */
	    gyro->snBit = 0;
	    gyro->snO = 1;
	 }
      }

      if (buf[2] & 0x01) gyro->sn |= gyro->snBit;
      gyro->snBit <<= 1;

      if (gyro->snO) {
	printf("gyro serial number : %d\n", gyro->sn);
      }
   }

   /* temperature */
   if (buf[4] & 0x4)
      gyro->tlsb = ((buf[4]<<5)&0x7F) + ((buf[5]&0x7F)>>2);
   else {
      gyro->tmsb = ((buf[4]<<5)&0x7F) + ((buf[5]&0x7F)>>2);

      d = ((((short)gyro->tmsb)<<7) + gyro->tlsb) &0xFFF;
      if (d & 0x800) d += 0xFFFFF000;

      gyro->temp = d * GYRO_TEMP_RESOLUTION;
   }

   /* rate */

#if VERBOSE>1
   printf("%02x %02x %02x", buf[5], buf[6], buf[7]);
#endif

   d = (((short)(buf[5]&0x3))<<14)+(((short)buf[6]&0x7F)<<7)+(buf[7]&0x7F);
   d &= 0xFFFF;
   if (d & 0x8000) d += 0xFFFF0000;
   gyro->w = d * GYRO_RATE_RESOLUTION;
#if VERBOSE>1
   printf(", rate=%04x %f rad/s", d, gyro->w);
#endif
   

   /* Offset correction */

   /* calibration on : estimation of wOffset on going */
   if (gyro->nbMeasuresOffset) {
     if (!totMeasuresOffset)
       totMeasuresOffset = gyro->nbMeasuresOffset;
     gyro->wOffset += gyro->w/totMeasuresOffset;
     gyro->nbMeasuresOffset--;
   }
   /* calibration off : use wOffset to correct w */
   else {
     gyro->w -= gyro->wOffset;
     totMeasuresOffset = 0;
   }

   /* Computes angle */
   if (gyro->prevDate)
     gyro->period = gyro->date - gyro->prevDate;
   else {
     gyro->resetDate = gyro->date;
     gyro->period = 0;
   }
   gyro->prevDate = gyro->date;
   gyro->theta += gyro->w * gyro->period/1000.0;
   gyro->theta = angleLimit(gyro->theta);

#if VERBOSE>0
   printf(", %5.4f deg/s, dt=%u ms, th=%5.4f deg\n", 
	  RAD_TO_DEG(gyro->w), gyro->period, RAD_TO_DEG(gyro->theta));
#endif

   return 0;
}

/* ----------------------------------------------------------------------
 *
 * getTime
 *
 *
 */

static u_long getTime (void)
{
  u_long msec;
#ifdef HAS_POSIX_CLOCK
  struct timespec tp; /*  mesure du temps */
#else
  struct timeval tv;
#endif /* HAS_POSIX_CLOCK */



#ifdef HAS_POSIX_CLOCK
  clock_gettime(CLOCK_REALTIME, &tp);
  msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
#else
  gettimeofday(&tv, NULL);
  msec = (tv.tv_usec / 1000) + (tv.tv_sec * 1000);
#endif /* HAS_POSIX_CLOCK */
  
  return msec;
}

