/*
 * Copyright (c) 2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/* ---------------------------------------------------------------------- 
 * Functions to init and read the angle from  a KVH5000 gyro
 * ---------------------------------------------------------------------- */

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HAS_POSIX_CLOCK
#include <time.h>
#else
#include <sys/time.h>
#endif

#include <portLib.h>
#include <h2mathLib.h>

#include "serial.h"
#include "../gyroLib.h"

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ((x)*3.1415926/180.)
#endif

#define DEBUG 0

/* ---------------------------------------------------------------------- */
static int fd;
static double angleInit;
static BOOL gyroReadAngleOnce(double *theta);
static BOOL gyroInit(char *port);

static u_long getTime()
{
  u_long res;
#ifdef HAS_POSIX_CLOCK
  struct timespec tp; /*  mesure du temps */
#else
  struct timeval tv;
#endif /* HAS_POSIX_CLOCK */

#ifdef HAS_POSIX_CLOCK
    clock_gettime(CLOCK_REALTIME, &tp);
    res = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
#else
    gettimeofday(&tv, NULL);
    res = (tv.tv_usec / 1000) + (tv.tv_sec * 1000);
#endif /* HAS_POSIX_CLOCK */

	return res;
}



/* ----------------------------------------------------------------------
 *
 * gyroInitAngle -  Init the serial port and switch to mode angle
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL gyroInitAngle(char *port)
{
  int nbw;

  /* init */
  if (!gyroInit(port)) {
    return FALSE;
  }

  
  if (gyroResetAngle(0.0) == FALSE) {
    close(fd);
    return FALSE;
  }

  /* switch to mode angle */
  if ((nbw = writeSerial(fd, "P", 1)) != 1) {
    fprintf(stderr, "writeSerial : %d written (1 expected) : %s\n", 
	    nbw, strerror(errno));
    close(fd);
    return FALSE;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * gyroEnd -     End connection to serial port
 *
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL gyroEnd(void)
{
  if (fd != -1) {
    if(close(fd) == -1) {
      printf ("gyroEnd: close(fd %d) failed\n", fd);
      return FALSE;
    }
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * gyroInit -     Init connection to serial port
 *
 *
 * Returns :            FALSE if failed or TRUE.
 */

static  u_long msec0;

static BOOL gyroInit(char *port)
{

  fprintf(stderr, "gyroInit %s\n", port);

  if ((fd = openSerial(port, 38400)) == -1) {
    fprintf(stderr, "serialInit : fd == -1: %s\n", strerror(errno));
    return FALSE;
  }
  if (!fd) {
    fprintf(stderr, "serialInit : fd == 0: %s\n", strerror(errno));
    return FALSE;
  }

  msec0 = getTime();

  return TRUE;
}


/* ----------------------------------------------------------------------
 *
 * gyroReadAngle -     Try to read the angle
 *                     using gyroReadAngleOnce().
 *
 *                      Try several times.
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL gyroReadAngle(double *theta)
{
  int nbTest=0;
  BOOL status;
  
  do {
    status = gyroReadAngleOnce(theta);
    if (!status)
      printf ("gyroReadAngleOnce: retry %d\n", ++nbTest);
	else  {
      // add the difference between the mesured angle and the angle Init 
      *theta += angleInit;	
      *theta = angleLimit(*theta);
	}

  } while(!status && nbTest<3);
  return status;
}

/* ----------------------------------------------------------------------
 *
 * gyroReadCorrAngle -     Like gyroReadAngle but correct the resulting 
 *                         angle taking into account the earth rotation from 
 *                         the gyroInit().
 *
 * WARNING : to be use only if the gyro is horizontal !
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL gyroReadCorrAngle(double *theta)
{
  BOOL status;
  u_long msec1;
  int duration;
  double corrAngle;

  status = gyroReadAngle(theta);
  if (status) {    
    msec1 = getTime();
    duration = msec1 - msec0;
    corrAngle = duration / 1000. * 2 * PI / 86400;
    *theta -= corrAngle;
    *theta = angleLimit(*theta);
  }
  return status;
}


#define NB_READ_CHAR 1024
#define KVH5000_STRING_LENGTH 21

// message of size KVH5000_STRING_LENGTH
static BOOL gyroReadLastMessage(char *message)
{
	char *stringEnd, *stringStart;
	int nbread;
	static char stringin[NB_READ_CHAR+1];
	static int sibegin = 0;
	int minread;
	
	while (1)
	{
		#if DEBUG>=8
		printf ("%s:%d\n", __FILE__, __LINE__);
		#endif
		
		minread = KVH5000_STRING_LENGTH-sibegin;
		if (minread < 0) minread = 0;
		
		nbread = readSerialMinMax(fd, stringin+sibegin, 
					minread, 
					NB_READ_CHAR-sibegin);
		
		#if DEBUG>=2
		printf ("gyroReadLastMessage: read %d char\n", nbread);
		#endif
		
		if (nbread == NB_READ_CHAR-sibegin)
		{
			stringin[sibegin+nbread] = 0;
			#if DEBUG>=3
			printf ("%s:%d stringin \"%s\"\n", __FILE__, __LINE__, stringin);
			#endif
			
			memcpy(stringin, stringin+NB_READ_CHAR-2*KVH5000_STRING_LENGTH, 2*KVH5000_STRING_LENGTH);
			sibegin = 2*KVH5000_STRING_LENGTH;
			continue;
		}
		stringin[sibegin+nbread] = 0;
		
		#if DEBUG>=3
		printf ("%s:%d stringin \"%s\"\n", __FILE__, __LINE__, stringin);
		#endif
		
		stringEnd = strrchr(stringin, '\n');
		if (stringEnd == 0) // we do have read KVH5000_STRING_LENGTH, so it's not normal
		{
			printf ("gyroReadLastMessage: bad format (no stringEnd in stringin)\n");
			printf ("nbread was %d, sibegin was %d\n", nbread, sibegin);
			printf ("stringin was \"%s\"\n", stringin);
			sibegin = 0;
			return FALSE;
		}
		if (stringEnd-stringin+1 < KVH5000_STRING_LENGTH) // the initial state was not ok, read next message
		{
printf ("%s:%d\n", __FILE__, __LINE__);
			sibegin = stringin+sibegin+nbread - stringEnd - 1;
			memmove(stringin, stringEnd+1, sibegin);
			continue;
		}
		if (stringEnd-stringin+1 > KVH5000_STRING_LENGTH)
		{
printf ("%s:%d\n", __FILE__, __LINE__);
			*stringEnd = 0;
			stringStart = strrchr(stringin, '\n')+1;
			if (stringEnd-stringStart+1 != KVH5000_STRING_LENGTH)
			{
				printf ("gyroReadLastMessage: bad format (string length %d)\n", stringEnd-stringStart+1);
				printf ("nbread was %d, sibegin was %d\n", nbread, sibegin);
				printf ("stringin was \"%s\"\n", stringin);
				sibegin = 0;
				return FALSE;
			}
		} else
			stringStart = stringin;
		
		memcpy(message, stringStart, KVH5000_STRING_LENGTH);
		sibegin = stringin+sibegin+nbread - stringEnd - 1;
		memmove(stringin, stringEnd+1, sibegin);
		return TRUE;
	}
}


/* ----------------------------------------------------------------------
 *
 * gyroReadAngleOnce - try to read and decode a string that should
 *                     contains the angle.
 *
 * Returns :            FALSE if failed or TRUE.
 */

/* We only need about 20 char but we read twice the  length to be 
   sure to get at least one complet message.... 

   ... and I do not know why but with 40 there are much more mistakes
   (wrong caracters) than with 64 !!!!

   Remark : with no taskDelay between two calls of gyroReadAngleOnce(), 
   there is no mistake.  Linux + serial = strange.
*/
#define KVH5000_POWER 6

static BOOL gyroReadAngleOnce(double *theta)
{
//	char *string, *stringEnd, *stringBegin;
	char string[KVH5000_STRING_LENGTH];
//	int stringLength;
//	int nbread;
	long int number1, number2, type;
	char *next, *start;
	int power;
	#if DEBUG>=1
	static int count;
	#endif

	/* --
	* -- read NB_READ_CHAR chars
	* -- */
	
	if (gyroReadLastMessage(string) == FALSE)
		return FALSE;
	
#if 0
	if ((nbread = readSerialMinMax(fd, stringin, 
				2*KVH5000_STRING_LENGTH, 
				NB_READ_CHAR)) != NB_READ_CHAR) {
		#if DEBUG>=2
		printf ("gyroReadOnce: read %d char\n", nbread);
		#endif
		if (!nbread)
			return FALSE;
	}
	stringin[nbread] = '\0';
	#if DEBUG>=3
	printf ("stringin %s\n", stringin);
	#endif

	/* --
	* -- look for last interesting string 
	* -- */

	/* look for the end of the string */
	stringEnd = strrchr(stringin, '\n');
	if (!stringEnd) {
		printf ("gyroReadOnce: bad format (no stringEnd in %s)\n", stringin);
		printf ("nbread was %d\n", nbread);
		printf ("stringin was \"%s\"\n", stringin);
		return FALSE;
	}
	*stringEnd = '\0';
	
	if (strlen(stringin) == KVH5000_STRING_LENGTH) {
		string = stringin;
	}
	else {
		/* look for the begining of the string */
		stringBegin = strrchr(stringin, '\n');
		if (!stringBegin) {
			printf ("gyroReadOnce: bad format (no stringBegin in %s)\n", stringin);
			printf ("nbread was %d\n", nbread);
			printf ("stringin was \"%s\"\n", stringin);
			return FALSE;
		}
		string = stringBegin;
	}
	/* --
	* --  decode
	* -- */

	#if DEBUG>=3
	printf ("string \"%s\"\n", string);
	#endif

	/* control format */
	stringLength = stringEnd - string;
	if (stringLength != KVH5000_STRING_LENGTH) {
		printf ("gyroReadOnce: bad format (string length %d)\n", stringLength);
		printf ("nbread was %d\n", nbread);
		printf ("stringin was \"%s\"\n", stringin);
		return FALSE;
	}
	#if DEBUG>=2
	printf ("string length = %d\n", stringLength);
	#endif
#endif

	/* first number part */
	number1 = strtol(string, &next, 10);
	#if DEBUG>=2
	printf ("number1 = %ld\n", number1);
	#endif

	/* control format */
	if (*next != '.') {
		printf ("gyroReadOnce: bad format (no '.')\n");
		printf ("string was \"%s\"\n", string);
		return FALSE;
	}

	/* second number part */
	start = next+1;
	number2 = strtol(start, &next, 10);
	power = next - start;
	#if DEBUG>=2
	printf ("number2 = %ld, power = %d\n", number2, power);
	#endif

	/* control format */
	if (power != KVH5000_POWER) {
		printf ("gyroReadOnce: bad format (power = %d)\n", power);
		printf ("string was \"%s\"\n", string);
		return FALSE;
	}
	
	/* control type */
	type = strtol(next+1, &next, 10);
	#if DEBUG>=2
	printf ("type = %ld\n", type);
	#endif
	if (type != 1) {
		printf ("gyroReadOnce: bad format (type %ld, not 1)\n", type);
		printf ("string was \"%s\"\n", string);
		return FALSE;
	}

	*theta = number1 + (double)number2/(double)pow(10,power);

	#if DEBUG>=2
	printf ("%ld . %ld = %f\n", number1, number2, *theta);
	#endif


	*theta = angleLimit(-DEG_TO_RAD(*theta));

	#if DEBUG>=1
	if (count==100) {
		printf ("gyro    : %6.3f deg = %7.4f rad\n", RAD_TO_DEG(*theta), *theta);
		count=0;
	}
	count++;
	#endif

	return TRUE;
}

// dummy implementation
BOOL
gyroUpdateWOffset()
{
	return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * gyroResetAngle - Reset the integrated angle to the value parameter
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL
gyroResetAngle(double angle)
{
	int nbw;

	/* Reset the Integrated Angle Value */
	if ((nbw = writeSerial(fd, "Z", 1)) != 1) {
		fprintf(stderr, "writeSerial : %d written (1 expected) : %s\n", 
			nbw, strerror(errno));
		return FALSE;
	}

	/*
	 * XXX : Workaround
	 * If we don't wait a few time before writing another command, it seems that
	 * the gyro just freezes, so ok, wait a second before doing any other thing
	 */
	sleep(1);

	angleInit = angle;
	msec0 = getTime();

	return TRUE;
}

